# -*- coding: utf-8 -*-
# Pretty-printers for TQt.

#  Copyright (C) 2013 Alexander Golubev (Fat-Zer) <fatzer2@gmail.com>
# 
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import gdb
import itertools
import re

#######################################################################
#                               Strings                               #
#######################################################################

class TQStringPrinter:

	def __init__(self, val):
		self.val = val

	def to_string(self):
		size = self.val['d']['len']
		try:
			dataAsCharPointer = self.val['d']['unicode'].cast(gdb.lookup_type("char").pointer())
			rv = dataAsCharPointer.string(encoding = 'UTF-16', length = size * 2)
		except Exception:
			pass
		return rv

	def display_hint (self):
		return 'string'

#######################################################################
#							Date and Time							#
#######################################################################
		
class TQDatePrinter:

	def __init__(self, val):
		self.val = val

	def to_string(self):
		julianDay = self.val['jd']

		if julianDay == 0:
			return "invalid TQDate"

		# Copied from Qt sources (Julian day for 1752-09-14)
		if julianDay < 2361222: 
			return "tqt doesn't operate on Julian-calendar dates (pre 1752-09-14)"
		
		# Gregorian calendar starting from October 15, 1582
		# This algorithm is from Henry F. Fliegel and Thomas C. Van Flandern
		ell = julianDay + 68569;
		n = (4 * ell) / 146097;
		ell = ell - (146097 * n + 3) / 4;
		i = (4000 * (ell + 1)) / 1461001;
		ell = ell - (1461 * i) / 4 + 31;
		j = (80 * ell) / 2447;
		d = ell - (2447 * j) / 80;
		ell = j / 11;
		m = j + 2 - (12 * ell);
		y = 100 * (n - 49) + i + ell;
		
		return "%d-%02d-%02d" % (y, m, d)

class TQTimePrinter:

	def __init__(self, val):
		self.val = val

	def to_string(self):
		ds = self.val['ds']

		if ds == -1:
			return "invalid TQTime"

		MSECS_PER_HOUR = 3600000
		SECS_PER_MIN = 60
		MSECS_PER_MIN = 60000

		hour = ds / MSECS_PER_HOUR
		minute = (ds % MSECS_PER_HOUR) / MSECS_PER_MIN
		second = (ds / 1000)%SECS_PER_MIN
		msec = ds % 1000
		return "%02d:%02d:%02d.%03d" % (hour, minute, second, msec)

class TQDateTimePrinter:

	def __init__(self, val):
		self.val = val

	def to_string(self):
		date = self.val['d'];
		time = self.val['t'];
		return "%s %s" % (self.val['d'], self.val['t'])


#######################################################################
#						  Container classes						  #
#######################################################################

class TQValueListPrinter:
	"Print a TQValueList"

	class _iterator:
		def __init__(self, nodetype, begin, size):
			self.nodetype = nodetype
			self.it = begin
			self.pos = 0
			self.size = size

		def __iter__(self):
			return self

		def next(self):
			if self.pos >= self.size:
				raise StopIteration

			pos = self.pos
			val = self.it['data']
			self.it = self.it['next']
			self.pos = self.pos + 1
			return ('[%d]' % pos, val)

	def __init__(self, val):
		self.val = val
		self.itype = self.val.type.template_argument(0)

	def children(self):
		return self._iterator(self.itype, self.val['sh']['node']['next'], self.val['sh']['nodes'])

	def to_string(self):
		if self.val['sh']['nodes'] == 0:
			empty = " [empty]"
		else:
			empty = ""

		return "TQValueList<%s>%s" % ( self.itype, empty )

class TQPtrListPrinter:
	"Print a TQPtrList"

	class _iterator:
		def __init__(self, nodetype, first):
			self.nodetype = nodetype
			self.it = first
			self.pos = 0

		def __iter__(self):
			return self

		def next(self):
			if self.it == 0:
				raise StopIteration
			pos = self.pos
			ptr = self.it['data'].cast(self.nodetype.pointer())
			try:
				rv = str(ptr.dereference())
			except RuntimeError:
				rv = "Cannot access object on address %#x" % ptr
				pass
 			
			self.it = self.it['next']
			self.pos = self.pos + 1

			return ('[%d]' % pos, rv)

	def __init__(self, val):
		self.val = val
		self.itype = self.val.type.template_argument(0)

	def children(self):
		return self._iterator(self.itype, self.val['firstNode'])

	def to_string(self):
		if self.val['firstNode'] == 0:
			empty = " [empty]"
		else:
			empty = ""

		return "TQPtrList<%s>%s" % ( self.itype, empty )

class TQStrListPrinter:
	"Print a TQStrList"

	class _iterator:
		def __init__(self, first):
			self.it = first
			self.pos = 0

		def __iter__(self):
			return self

		def next(self):
			if self.it == 0:
				raise StopIteration
			pos = self.pos
			ptr = self.it['data'].cast(gdb.lookup_type('char').pointer())
			try:
				rv = str(ptr)
			except RuntimeError:
				# souldn't happen for QStrList
				rv = "Cannot access c-string on address %#x" % ptr
				pass
 			
			self.it = self.it['next']
			self.pos = self.pos + 1

			return ('[%d]' % pos, rv)

	def __init__(self, val):
		self.val = val

	def children(self):
		return self._iterator(self.val['firstNode'])

	def to_string(self):
		if self.val['firstNode'] == 0:
			empty = " [empty]"
		else:
			empty = ""

		return "TQStrList%s" % ( empty )

def tqt_lookup_function (val):
	"Look-up and return a pretty-printer that can print val."
	
	# Get the type.
	type=val.type
	# If it points to a reference, get the reference.
	if type.code == gdb.TYPE_CODE_REF:
		type = type.target ()
	# Get the unqualified type, stripped of typedefs.
	type = type.unqualified ().strip_typedefs ()


	# Get the type name.
	typename = type.tag
	if typename == None:
		return None
	
	# Iterate over local dictionary of types to determine
	# if a printer is registered for that type.  Return an
	# instantiation of the printer if found.
	for function in tqt_pretty_printers_dict:
		if function.search (typename):
			return tqt_pretty_printers_dict[function] (val)

	# Cannot find a pretty printer.  Return None.
	return None

def tqt_build_dictionary ():
	tqt_pretty_printers_dict[re.compile('^TQString$')] = lambda val: TQStringPrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQDate$')] = lambda val: TQDatePrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQTime$')] = lambda val: TQTimePrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQDateTime$')] = lambda val: TQDateTimePrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQValueList<.*>$')] = lambda val: TQValueListPrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQPtrList<.*>$')] = lambda val: TQPtrListPrinter(val)
	tqt_pretty_printers_dict[re.compile('^TQStrList$')] = lambda val: TQStrListPrinter(val)

tqt_pretty_printers_dict = {}
tqt_build_dictionary ()

def register_tqt3_printers (obj):
	if obj == None:
		obj = gdb
	obj.pretty_printers.append (tqt_lookup_function)
